/*  Monkey HTTP Daemon
 *  ------------------
 *  Copyright (C) 2001-2003, Eduardo Silva P.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>

#include "monkey.h"

/* Get & Head Method */
int M_METHOD_Get_and_Head(struct request *sr)
{
	/* *sr = struct request */
	char *location=0, *real_location=0; /* ruta para redireccion */
	char **arg_script=0, **ptr_script=0, **mime_info;

	struct stat checkpath;
	struct stat sizefile;

	/* Peticion normal, no es a un Virtualhost */
	if((strcmp(sr->uri_processed,"/"))==0)
		snprintf(sr->real_path, MAX_PATH, "%s", sr->temp_path);

	if(sr->user_home==VAR_OFF)
		snprintf(sr->real_path, MAX_PATH,"%s%s", sr->temp_path, sr->uri_processed);
	
	if(stat(sr->real_path, &checkpath)==-1){
		int method_value=0;
		if(sr->method!=HEAD_METHOD)
			method_value=1;

		Request_Error(M_CLIENT_NOT_FOUND, sr, method_value, sr->log);
		return -1;
	}

	/* Checkeando si la ruta es un Directorio */
	if(checkpath.st_mode & S_IFDIR) {
		int res_Findex=-1;
		/* Comprobamos si tiene un slash final,
 	 	si no lo tiene se envia una redireccion */
		if(sr->uri_processed[strlen(sr->uri_processed) - 1] != '/') {
			location=m_build_buffer("%s/", sr->uri_processed);
			if(config->serverport == config->standard_port)
				real_location=m_build_buffer("http://%s%s", sr->host, location);
			else
				real_location=m_build_buffer("http://%s:%i%s", sr->host, config->serverport, location);

			sr->headers->status = M_REDIR_MOVED;
			sr->headers->content_length = 0;
			sr->headers->content_type = NULL;
			sr->headers->location = real_location;
			sr->headers->cgi = SH_NOCGI;
			sr->headers->pconnections_left = config->max_keep_alive_request - sr->counter_connections;

			M_METHOD_send_headers(sr->socket, sr->headers, sr->log);

			M_free(location);
			M_free(real_location);
			
			return 0;
		}

		if(sr->forcegetdir==VAR_OFF)
			res_Findex=FindIndex(sr->real_path);

		/* Buscar un indexfile segn configuracin */
		if(res_Findex!=0) {
			/* No existe un indexfile, mostrar el directorio */
			if(config->getdir==VAR_ON) {
				if(GetDir(sr->real_path, sr->socket, sr->uri_processed, sr->log, sr->protocol) == -1)
					Request_Error(M_CLIENT_FORBIDDEN, sr, 1, sr->log);

				return -1;
			}
			else {
				Request_Error(M_CLIENT_NOT_FOUND, sr, 1, sr->log);
				return -1;
			}
		}
	}
	/* Existe archivo? */
	if(access(sr->real_path,F_OK)!=0){
		Request_Error(M_CLIENT_NOT_FOUND, sr, 1, sr->log);
		return -1;	
	}

	/* Permisos de lectura */
	if(AccessFile(sr->real_path)!=0){
		Request_Error(M_CLIENT_FORBIDDEN, sr, 1, sr->log);
		return -1;	
	}
		
	/* Buscanco MimeType que coincida*/
	mime_info=(char **) alloca(sizeof(char *) * 3);
	mime_info=Mimetype_Find(sr->real_path);
	
	if(mime_info[1]){
	/*  Script ejecutable (e.g PHP) ? */
	if(access(mime_info[1],X_OK)==0){
			int cgi_status=0;

			/*  Es un archivo regular ? */
			if(CheckFile(mime_info[1])!=0){
				Request_Error(M_SERVER_INTERNAL_ERROR, sr, 1, sr->log);
				return -1;
			}

			sr->log->final_response=M_HTTP_OK;
			sr->script_filename=M_strdup(sr->real_path);
			arg_script = ptr_script = (char **) alloca(sizeof(char *) * 3);
			*ptr_script++ = M_strdup(mime_info[1]);
			*ptr_script++ = M_strdup(sr->script_filename);
			*ptr_script++ = '\0';

			if(sr->method==GET_METHOD){
					cgi_status=M_CGI_run_get(sr, mime_info[1], arg_script);
			}
			else if(sr->method==POST_METHOD) {
					cgi_status=M_CGI_run_post(sr, mime_info[1], arg_script);
			}
			
			switch(cgi_status){
				case -2:	/* Timeout */
						sr->log->final_response=M_CLIENT_REQUEST_TIMEOUT;
						break;
				case -3:  /* Internal server Error */
						sr->log->final_response=M_SERVER_INTERNAL_ERROR;
						break;
						
				case -1:
						sr->make_log=VAR_OFF;
						break;
				case 0:  /* Ok */
						sr->log->final_response=M_HTTP_OK;
						break;
			};	

			if(cgi_status==M_CGI_TIMEOUT || cgi_status==M_CGI_INTERNAL_SERVER_ERR){
				Request_Error(sr->log->final_response, sr, 1, sr->log);	
			}
			arg_script = ptr_script = NULL;
			return cgi_status;
	}
	}
	/* Rescatando largo del archivo */
	if(stat(sr->real_path,&sizefile) < 0) {
		Request_Error(M_CLIENT_NOT_FOUND, sr, 1, sr->log);
		return -1;
	}
	
	/* Enviando cabeceras */
	sr->headers->status = M_HTTP_OK;
	sr->headers->content_length = sizefile.st_size;
	sr->headers->cgi = SH_NOCGI;
	sr->headers->location = NULL;
	sr->headers->pconnections_left = config->max_keep_alive_request - sr->counter_connections;

	sr->log->size=(sizefile.st_size);
	if(sr->method==GET_METHOD || sr->method==POST_METHOD){
		sr->headers->content_type = mime_info[0];
		/* Range */
		if(sr->range!=NULL && config->resume==VAR_ON){
			M_METHOD_get_range(sr->range, sr->headers->range_values);
			if(sr->headers->range_values[0]>=0 || sr->headers->range_values[1]>=0)
				sr->headers->status = M_HTTP_PARTIAL;
		}
	}
	else{ /* Sin content-type */
		sr->headers->content_type = NULL;
	}
	M_METHOD_send_headers(sr->socket, sr->headers, sr->log);

	/* Enviando archivo */
	if(sr->method==GET_METHOD || sr->method==POST_METHOD)
		SendFile(sr->socket, sr->real_path, sr->headers->range_values);

	free(mime_info);
	return 0;
}


/* POST METHOD */
int M_METHOD_Post(struct request *s_request, char *request_body)
{
	char *post_buffer;
	char buffer[MAX_REQUEST_BODY];
	int i=0, content_length_post=0;
	
	if(!(Request_Find_Variable(request_body, RH_CONTENT_LENGTH))){
		Request_Error(M_CLIENT_LENGHT_REQUIRED, s_request,0,s_request->log);
		return -1;
	}
	
	content_length_post = (int) atoi(Request_Find_Variable(request_body, RH_CONTENT_LENGTH));

	if(content_length_post<=0 || content_length_post >=MAX_REQUEST_BODY){
		Request_Error(M_CLIENT_BAD_REQUEST ,s_request, 0, s_request->log);	
		return -1;
	}
	
	if(!(Request_Find_Variable(request_body, RH_CONTENT_TYPE))){
		Request_Error(M_CLIENT_BAD_REQUEST, s_request, 0, s_request->log);
		return -1;
	}
	
	s_request->content_type = M_strdup(Request_Find_Variable(request_body, RH_CONTENT_TYPE));
	post_buffer = (char *) strstr(request_body,"\r\n\r\n");

	if(post_buffer==NULL || strlen(post_buffer)<=4) {
		s_request->post_variables=NULL;
		return -1;
	}

	memset(buffer,'\0',sizeof(buffer));
	for(i=4;i<strlen(post_buffer);i++){
		buffer[i-4]=post_buffer[i];
	}

	if(strlen(buffer) < content_length_post){
		content_length_post=strlen(buffer);
	}
			
	s_request->post_variables = M_malloc(sizeof(buffer) + 1);
	memset(s_request->post_variables, '\0', sizeof(s_request->post_variables));
	strncpy(s_request->post_variables, buffer, content_length_post);
	s_request->post_variables[content_length_post ]='\0';
	s_request->content_length=content_length_post;
	return 0;
}

/* Send_Header , envia las cabeceras principales */
int M_METHOD_send_headers(int fd, struct header_values *sh, struct log_info *s_log)
{

	int fd_status=0;
	/* Status Code */
	switch(sh->status){
		case M_HTTP_OK:	
			s_log->final_response=M_HTTP_OK;
			fd_status = fdprintf(fd, NO_CHUNKED, "HTTP/1.1 200 OK\r\n");
			break;
			
		case M_HTTP_PARTIAL:	
			s_log->final_response=M_HTTP_PARTIAL;
			fd_status = fdprintf(fd, NO_CHUNKED, "HTTP/1.1 206 Partial Content\r\n");
			break;
			
		case M_REDIR_MOVED:
			s_log->final_response=M_REDIR_MOVED;
			s_log->status=S_LOG_OFF;
			fd_status = fdprintf(fd, NO_CHUNKED, "HTTP/1.1 301 Moved Permanently\r\n");
			break;

		case M_REDIR_MOVED_T:
			s_log->final_response=M_REDIR_MOVED_T;
			s_log->status=S_LOG_ON;
			fd_status = fdprintf(fd, NO_CHUNKED, "HTTP/1.1 302 Found\r\n");
			break;
	};

	if(fd_status<0){
		return -1;		
	}
	
	/* Informacion del server */
	fdprintf(fd, NO_CHUNKED, "Server: %s\r\n", config->server_software);
	
	/* Fecha */
	fdprintf(fd, NO_CHUNKED, "Date: %s\r\n",PutDateHeader(0));
	
	/* Location */
	if(sh->location!=NULL)
		fdprintf(fd, NO_CHUNKED, "Location: %s\r\n", sh->location);

	/* Conexion */
	if(sh->pconnections_left!=0 && config->keep_alive==VAR_ON){
		fdprintf(fd, NO_CHUNKED, "Keep-Alive: timeout=%i, max=%i\r\n", config->keep_alive_timeout, sh->pconnections_left);
		fdprintf(fd, NO_CHUNKED, "Connection: Keep-Alive\r\n");
	}
	else
		fdprintf(fd, NO_CHUNKED, "Connection: close\r\n");

	/* Tipo de contenido de la informacion */
	if(sh->content_type!=NULL){
		fdprintf(fd, NO_CHUNKED, "Content-Type: %s\r\n", sh->content_type);
	}
	
	/* Ranges */ 
	fdprintf(fd, NO_CHUNKED, "Accept-Ranges: bytes\r\n", sh->content_type);
		
	/* Tamao total de la informacion a enviar */
	if((sh->content_length!=0 && (sh->range_values[0]>=0 || sh->range_values[1]>=0)) && config->resume==VAR_ON){
		long int length;

		/* yyy- */
		if(sh->range_values[0]>=0 && sh->range_values[1]==-1){
			length = (unsigned long int) ( sh->content_length - sh->range_values[0] );
			fdprintf(fd, NO_CHUNKED, "%s %i\r\n", RH_CONTENT_LENGTH, length);
			fdprintf(fd, NO_CHUNKED, "%s bytes %d-%d/%d\r\n", RH_CONTENT_RANGE, sh->range_values[0], 
				(sh->content_length - 1), sh->content_length);
		}
		
		/* yyy-xxx */
		if(sh->range_values[0]>=0 && sh->range_values[1]>=0){
			length = (unsigned long int) (sh->range_values[1] - sh->range_values[0] + 1);
			fdprintf(fd, NO_CHUNKED, "%s %d\r\n", RH_CONTENT_LENGTH, length);	
			fdprintf(fd, NO_CHUNKED, "%s bytes %d-%d/%d\r\n", RH_CONTENT_RANGE, sh->range_values[0], 
				sh->range_values[1], sh->content_length);
		}

		/* -xxx */
		if(sh->range_values[0]==-1 && sh->range_values[1]>=0){
			fdprintf(fd, NO_CHUNKED, "%s %d\r\n", RH_CONTENT_LENGTH, sh->range_values[1]);
			fdprintf(fd, NO_CHUNKED, "%s bytes %d-%d/%d\r\n", RH_CONTENT_RANGE, (sh->content_length - sh->range_values[1]),
				 (sh->content_length - 1) , sh->content_length);
		}
	}
	else if(sh->content_length!=0)
		fdprintf(fd, NO_CHUNKED, "%s %d\r\n", RH_CONTENT_LENGTH, sh->content_length);
		
	if(sh->cgi==SH_NOCGI)
		fdprintf(fd, NO_CHUNKED, "\r\n");

	return 0;
}

int M_METHOD_get_number(char *method)
{
	if(strcmp(method,"GET")==0)
		return GET_METHOD;

	if(strcmp(method,"POST")==0)
		return POST_METHOD;

	if(strcmp(method,"HEAD")==0)
		return HEAD_METHOD;

	return METHOD_NOT_ALLOWED;
}

char *M_METHOD_get_name(int method)
{
	switch(method){
		case GET_METHOD:
				return (char *) "GET";
				
		case POST_METHOD:
				return (char *) "POST";
				
		case HEAD_METHOD:
				return (char *) "HEAD";
	}
	return (char *) "";
}

int M_METHOD_get_range(char *header, int range_from_to[2])
{
	int eq_pos, sep_pos;
	
	range_from_to[0] = -1;
	range_from_to[1] = -1;
	
	if(!header)
		return -1;	
	
	if((eq_pos = str_search(header, "=", 1))<0)
		return -1;	
	
	if(strncasecmp(header, "Bytes", eq_pos)!=0)
		return -1;	
	
	if((sep_pos = str_search(header, "-", 1))<0)
		return -1;
	
	/* =-xxx */
	if(eq_pos+1 == sep_pos){
		range_from_to[0] = -1;
		range_from_to[1] = (unsigned long) atol(header + sep_pos + 1);
		return 0;
	}

	/* =yyy-xxx */
	if( (eq_pos+1 != sep_pos) && (strlen(header) > sep_pos + 1) ){
		char *buffer=0;
		
		buffer = M_strdup(header+eq_pos+1);
		buffer = strtok(buffer, "-");
		range_from_to[0] = (unsigned long) atol(m_build_buffer("%d", atol(buffer)));
		buffer = strtok(NULL, "\n");
		range_from_to[1] = (unsigned long) atol(m_build_buffer("%d", atol(buffer)));	
		free(buffer);
		return 0;
	}
	/* =yyy- */
	if( (eq_pos+1 != sep_pos) && (strlen(header) == sep_pos + 1 ) ){
		char *buffer=0;

		buffer = M_strdup(header+eq_pos+1);
		buffer = strtok(buffer, "-");

		range_from_to[0] = (unsigned long) atol(m_build_buffer("%d", atol(buffer)));
		range_from_to[1] = -1;
		free(buffer);		
		return 0;
	}
	
	return -1;	
}
